@@ -1,5 +1,5 @@
 ###############################################################################
-# Copyright 2017-2021 - Climate Research Division
+# Copyright 2017-2023 - Climate Research Division
 #                       Environment and Climate Change Canada
 #
 # This file is part of the "fstd2nc" package.
@@ -23,7 +23,7 @@
 Functionality for converting between FSTD and netCDF files.
 """
 
-__version__ = "0.20230328.0"
+__version__ = "0.20231105.2"
 
 
 # Check for bundled rpnpy package.
@@ -52,6 +52,7 @@ from fstd2nc.mixins.select import SelectVars
 from fstd2nc.mixins.ascii import ASCII
 from fstd2nc.mixins.masks import Masks
 from fstd2nc.mixins.dates import Dates
+from fstd2nc.mixins.accum import Accum
 from fstd2nc.mixins.ensembles import Ensembles
 from fstd2nc.mixins.series import Series
 from fstd2nc.mixins.sfc_codes import Sfc_Codes
@@ -69,7 +70,7 @@ from fstd2nc.mixins.compat import FSTD_Compat
 from fstd2nc.mixins.extern import ExternInput, ExternOutput
 from fstd2nc.mixins.diaghacks import DiagHacks
 
-class Buffer (DiagHacks,ExternOutput,FSTD_Compat,netCDF_IO,netCDF_Atts,PruneAxes,Crop,YinYang,Interp,GridHacks,RemoveStuff,FilterRecords,NoNK,Mesh,XYCoords,VCoords,Sfc_Codes,VarDict,Series,Ensembles,Dates,Masks,ASCII,SelectVars,ExternInput,FSTD):
+class Buffer (DiagHacks,ExternOutput,FSTD_Compat,netCDF_IO,netCDF_Atts,PruneAxes,Crop,YinYang,Interp,GridHacks,RemoveStuff,FilterRecords,NoNK,Mesh,XYCoords,VCoords,Sfc_Codes,VarDict,Series,Ensembles,Accum,Dates,Masks,ASCII,SelectVars,ExternInput,FSTD):
   """
   High-level interface for FSTD data, to treat it as multi-dimensional arrays.
   Contains logic for dealing with most of the common FSTD file conventions.
