@@ -1,5 +1,5 @@
 ###############################################################################
-# Copyright 2017-2021 - Climate Research Division
+# Copyright 2017-2023 - Climate Research Division
 #                       Environment and Climate Change Canada
 #
 # This file is part of the "fstd2nc" package.
@@ -164,7 +164,6 @@ def decode_headers (raw):
   out['nj'] = np.empty(nrecs, dtype='int32')
   out['datyp'] = np.empty(nrecs, dtype='ubyte')
   out['nk'] = np.empty(nrecs, dtype='int32')
-  out['ubc'] = np.empty(nrecs, dtype='uint16')
   out['npas'] = np.empty(nrecs, dtype='int32')
   out['ig1'] = np.empty(nrecs, dtype='int32')
   out['ig2'] = np.empty(nrecs, dtype='int32')
@@ -177,10 +176,6 @@ def decode_headers (raw):
   out['ip2'] = np.empty(nrecs, dtype='int32')
   out['ip3'] = np.empty(nrecs, dtype='int32')
   out['datev'] = np.empty(nrecs, dtype='int32')
-  out['dateo'] = np.empty(nrecs, dtype='int32')
-  out['xtra1'] = np.empty(nrecs, dtype='uint32')
-  out['xtra2'] = np.empty(nrecs, dtype='uint32')
-  out['xtra3'] = np.empty(nrecs, dtype='uint32')
 
   temp8 = np.empty(nrecs, dtype='ubyte')
   temp32 = np.empty(nrecs, dtype='int32')
@@ -192,7 +187,7 @@ def decode_headers (raw):
   np.divmod(raw[:,1,0],256, out['deet'], out['nbits'])
   np.divmod(raw[:,1,1],256, out['ni'], out['grtyp'].view('ubyte'))
   np.divmod(raw[:,2,0],256, out['nj'], out['datyp'])
-  np.divmod(raw[:,2,1],4096, out['nk'], out['ubc'])
+  np.divmod(raw[:,2,1],4096, out['nk'], temp32)
   out['npas'][:] = raw[:,3,0]//64
   np.divmod(raw[:,3,1],256, out['ig4'], temp32)
   out['ig2'][:] = (temp32 << 16) # ig2a
@@ -204,38 +199,49 @@ def decode_headers (raw):
   etik6a = raw[:,5,1]//4
   et = raw[:,6,0]//256
   etikbc, _typvar = divmod(et, 4096)
+  del et
   _nomvar = raw[:,6,1]//256
   np.divmod(raw[:,7,0],16, out['ip1'], temp8)
   out['ip2'][:] = raw[:,7,1]//16
   out['ip3'][:] = raw[:,8,0]//16
-  date_stamp = raw[:,8,1]
+  date_stamp = raw[:,8,1].astype('int32')
   # Reassemble and decode.
   # (Based on fstd98.c)
   etiket_bytes = np.empty((nrecs,12),dtype='ubyte')
   for i in range(5):
     etiket_bytes[:,i] = ((etik15 >> ((4-i)*6)) & 0x3f) + 32
+  del etik15
   for i in range(5,10):
     etiket_bytes[:,i] = ((etik6a >> ((9-i)*6)) & 0x3f) + 32
+  del etik6a
   etiket_bytes[:,10] = ((etikbc >> 6) & 0x3f) + 32
   etiket_bytes[:,11] = (etikbc & 0x3f) + 32
+  del etikbc
   out['etiket'][:] = etiket_bytes.flatten().view('|S12')
+  del etiket_bytes
   nomvar_bytes = np.empty((nrecs,4),dtype='ubyte')
   for i in range(4):
     nomvar_bytes[:,i] = ((_nomvar >> ((3-i)*6)) & 0x3f) + 32
+  del _nomvar
   out['nomvar'][:] = nomvar_bytes.flatten().view('|S4')
+  del nomvar_bytes
   typvar_bytes = np.empty((nrecs,2),dtype='ubyte')
   typvar_bytes[:,0] = ((_typvar >> 6) & 0x3f) + 32
   typvar_bytes[:,1] = ((_typvar & 0x3f)) + 32
+  del _typvar
   out['typvar'][:] = typvar_bytes.flatten().view('|S2')
-  out['datev'][:] = (date_stamp >> 3) * 10 + (date_stamp & 0x7)
-  # Note: this dateo calculation is based on my assumption that
-  # the raw stamps increase in 5-second intervals.
-  # Doing it this way to avoid a gazillion calls to incdat.
-  date_stamp = date_stamp - (out['deet']*out['npas'])//5
-  out['dateo'][:] = (date_stamp >> 3) * 10 + (date_stamp & 0x7)
-  out['xtra1'][:] = out['datev']
-  out['xtra2'][:] = 0
-  out['xtra3'][:] = 0
+  del typvar_bytes
+  # Convert raw stamp to RPN date code.
+  # Two cases: positive = regular, negative = extended range.
+  out['datev'][:] = np.where (date_stamp >= 0,
+    (date_stamp >> 3) * 10 + (date_stamp & 0x7),
+    ((date_stamp+858993488) >> 3) * 10 + ((date_stamp+858993488) & 0x7) - 6
+  )
+  del date_stamp
+  # NOTE: Date of origin is now calculated in the dates mixin, since it's more
+  # involved than a simple arithmetic operation.
+  # NOTE: xtra1, xtra2, xtra3 not returned (not used, and take up space in the
+  # header table).
   return out
 
 
@@ -326,6 +332,7 @@ def get_crs (dataset):
   """
   from fstd2nc.stdout import _, info, warn, error
   import cartopy.crs as ccrs
+  from math import asin, pi
   # Find the grid mapping variable in the dataset.
   gmap = set(varname for varname,var in dataset.variables.items() if 'grid_mapping_name' in var.attrs)
   # Must have exactly one unique grid projection in the Dataset.
@@ -336,15 +343,23 @@ def get_crs (dataset):
     warn(_("Multiple grid mappings found: %s")%gmap)
     return None
   gmap = dataset[gmap.pop()]
+  # Define the spherical globe.
+  radius = gmap.attrs.get('earth_radius',None)
+  globe = ccrs.Globe(ellipse="sphere",semimajor_axis=radius,semiminor_axis=radius)
   gname = gmap.attrs['grid_mapping_name']
   if gname == 'latitude_longitude':
-    proj = ccrs.PlateCarree()
+    proj = ccrs.PlateCarree(globe=globe)
   elif gname == 'rotated_latitude_longitude':
-    proj = ccrs.RotatedPole (pole_longitude = gmap.attrs['grid_north_pole_longitude'], pole_latitude = gmap.attrs['grid_north_pole_latitude'])
+    proj = ccrs.RotatedPole (pole_longitude = gmap.attrs['grid_north_pole_longitude'], pole_latitude = gmap.attrs['grid_north_pole_latitude'], central_rotated_longitude=gmap.attrs.get('north_pole_grid_longitude',0.), globe=globe)
   elif gname == 'polar_stereographic':
     # May fail for South Pole projections (would maybe need to change sign of
     # true_scale_latitude?)
-    proj = ccrs.Stereographic (central_latitude = gmap.attrs['latitude_of_projection_origin'], central_longitude = gmap.attrs['straight_vertical_longitude_from_pole'], false_easting = gmap.attrs['false_easting'], false_northing = gmap.attrs['false_northing'], true_scale_latitude=60)
+    if 'scale_factor_at_projection_origin' in gmap.attrs:
+      k = gmap.attrs['scale_factor_at_projection_origin']
+      lat = abs(asin(2*k-1)) / pi * 180
+    else:
+      lat = gmap.attrs['standard_parallel']
+    proj = ccrs.Stereographic (central_latitude = gmap.attrs['latitude_of_projection_origin'], central_longitude = gmap.attrs['straight_vertical_longitude_from_pole'], false_easting = gmap.attrs['false_easting'], false_northing = gmap.attrs['false_northing'], true_scale_latitude=lat, globe=globe)
   else:
     warn(_("Unhandled grid mapping: %s")%gname)
     return None
