@@ -1,5 +1,5 @@
 ###############################################################################
-# Copyright 2017-2021 - Climate Research Division
+# Copyright 2017-2023 - Climate Research Division
 #                       Environment and Climate Change Canada
 #
 # This file is part of the "fstd2nc" package.
@@ -49,6 +49,11 @@ def vectorize (f):
     return cache[x]
   @wraps(f)
   def vectorized_f (x):
+    mask = None
+    if hasattr(x,'mask'):
+      mask = np.ma.getmaskarray(x)
+      n = len(x)
+      x = x[~mask]
     if _use_pandas():
       from pandas import Series, unique
       # If we're given a scalar value, then simply return it.
@@ -60,12 +65,18 @@ def vectorize (f):
       outputs = map(f,inputs)
       table = dict(zip(inputs,outputs))
       result = Series(x).map(table)
-      return result.values
+      result = result.values
     else:
       # If we're given a scalar value, then simply return it.
       if not hasattr(x,'__len__'):
         return cached_f(x)
-      return list(map(cached_f,x))
+      result = list(map(cached_f,x))
+    if mask is not None:
+      result = np.asarray(result)
+      masked_result = np.ma.masked_all(n,dtype=result.dtype)
+      masked_result[~mask] = result
+      result = masked_result
+    return result
   return vectorized_f
 
 
@@ -130,24 +141,15 @@ class _iter_type (_base_type):
 # Similar to above, but more general.
 # Can allow for multiple records along the inner axes.
 class _chunk_type (_base_type):
-  __slots__ = ('name','atts','axes','dtype','chunks','chunksize','deps')
-  def __init__ (self, name, atts, axes, dtype, chunks, chunksize):
+  __slots__ = ('name','atts','axes','dtype','chunks','record_id','deps')
+  def __init__ (self, name, atts, axes, dtype, chunks, record_id):
     self.name = name
     self.atts = atts
     self.axes = axes
     self.dtype = dtype
     self.chunks = chunks
-    self.chunksize = chunksize
+    self.record_id = record_id
     self.deps = []
-  def items(self):
-    for key, value in self.chunks.items():
-      key = tuple(k if isinstance(k,int) else slice(*k) for k in key)
-      yield key, value
-  def keys(self):
-    for key, value in self.items():
-      yield key
-  def values(self):
-    return self.chunks.values()
 
 # Fake progress bar - does nothing.
 class _FakeBar (object):
@@ -155,6 +157,7 @@ class _FakeBar (object):
   def iter(self, it):
     for i in it: yield i
   def __next__(self): pass
+  def next(self): pass
   def finish(self): pass
 
 # Try importing progress module.
@@ -202,6 +205,59 @@ try:
 except ImportError:
   _ProgressBar = _FakeBar
 
+# Special version of dictionary that keeps track of unaccessed keys.
+class AccessCountDict (object):
+  def __init__ (self, **items):
+    from collections import Counter
+    self._dict = dict(items)
+    self._counter = Counter()
+  def __getitem__ (self, key):
+    self._counter[key] += 1
+    return self._dict[key]
+  def __setitem__ (self, key, value):
+    self._counter[key] += 1
+    self._dict[key] = value
+  def __getattr__ (self, name): return getattr(self._dict, name)
+  def __str__ (self): return self._dict.__str__()
+  def __repr__ (self): return self._dict.__repr__()
+
+def _expand_files (filename):
+  """
+  Expands the given filename/directory/glob/Path/etc. into an explicit list
+  of matching files.
+
+  Parameters
+  ----------
+  filename : str or Path or iterable
+      The thing to expand.
+
+  Returns
+  -------
+  List of matches, each match a tuple of (pattern,[files])
+  """
+  from pathlib import Path
+  from glob import glob
+  import os
+  if isinstance(filename,(str,Path)):
+    infiles = [filename]
+  else:
+    infiles = list(filename)
+
+  # Apply wildcard and directory expansion to filenames.
+  expanded_infiles = []
+  for infile in infiles:
+    if isinstance(infile,Path):
+      infile = str(infile)
+    for f in sorted(glob(infile)) or [infile]:
+      if os.path.isdir(f):
+        for dirpath, dirnames, filenames in os.walk(f,followlinks=True):
+          for filename in filenames:
+            expanded_infiles.append((infile,os.path.join(dirpath,filename)))
+      else:
+        expanded_infiles.append((infile,f))
+  return expanded_infiles
+
+
 # Define a class for encoding / decoding the data.
 # Each step is placed in its own "mixin" class, to make it easier to patch in 
 # new behaviour if more exotic files are encountered in the future.
@@ -209,11 +265,15 @@ class BufferBase (object):
 
   # Names of records that should be kept separate (never grouped into
   # multidimensional arrays).
-  _meta_records = ()
+  @classmethod
+  def _meta_records(cls):
+    return ()
   # Other records that should also be kept separate, but only if they are
   # 1-dimensional.  If they are 2D, then they should be processed as a normal
   # variable.
-  _maybe_meta_records = ()
+  @classmethod
+  def _maybe_meta_records(cls):
+    return ()
 
   # Attributes which could potentially be used as outer axes.
   # The values from the attribute will become the axis values.
@@ -243,6 +303,21 @@ class BufferBase (object):
   # (They're either internal to the file, or part of the data, not metadata).
   _ignore_atts = ('file_id','name','address','length','dtype','selected')
 
+  # Header columns which contain extra data that should be passed to _decode.
+  # Tuple of (offset, length, d) where:
+  # - offset is the address in the file where the data is located.
+  # - length is how much data to read.
+  # - d is an alternate source (dask graph?) for case where data is not
+  #   stored in a file.
+  _decoder_data = (('data',('address','length','d')),)
+
+  # Extra arguments to pull from columns of the table.
+  _decoder_extra_args = ()
+
+  # Extra (scalar) arguments needed for the decoder at runtime.
+  def _decoder_scalar_args (self):
+    return {}
+
   # Define any command-line arguments for reading FSTD files.
   @classmethod
   def _cmdline_args (cls, parser):
@@ -310,10 +385,9 @@ class BufferBase (object):
     """
     from collections import Counter
     import numpy as np
-    from glob import glob, has_magic
+    from glob import has_magic
     import os
     from multiprocessing import Pool
-    from pathlib import Path
     try:
       from itertools import imap  # Python 2
     except ImportError:
@@ -331,23 +405,7 @@ class BufferBase (object):
     else:
       existing_buffer = None
 
-    if isinstance(filename,(str,Path)):
-      infiles = [filename]
-    else:
-      infiles = list(filename)
-
-    # Apply wildcard and directory expansion to filenames.
-    expanded_infiles = []
-    for infile in infiles:
-      if isinstance(infile,Path):
-        infile = str(infile)
-      for f in sorted(glob(infile)) or [infile]:
-        if os.path.isdir(f):
-          for dirpath, dirnames, filenames in os.walk(f,followlinks=True):
-            for filename in filenames:
-              expanded_infiles.append((infile,os.path.join(dirpath,filename)))
-        else:
-          expanded_infiles.append((infile,f))
+    expanded_infiles = _expand_files (filename)
 
     # How to handle internal metadata.
     internal_metadata = kwargs.pop('internal_metadata',None)
@@ -380,20 +438,27 @@ class BufferBase (object):
       error(_("Unexpected arguments: %s"%(kwargs.keys())))
 
     # Extract headers from the files.
-    if len(expanded_infiles) == 1: Bar = _FakeBar
-    bar = Bar(_("Inspecting input files"), suffix='%(percent)d%% (%(index)d/%(max)d)', max=len(expanded_infiles))
+    if hasattr(progress,'next'):
+      bar = progress
+    else:
+      if len(expanded_infiles) == 1: Bar = _FakeBar
+      bar = Bar(_("Inspecting input files"), suffix='%(percent)d%% (%(index)d/%(max)d)', max=len(expanded_infiles))
 
     if len(expanded_infiles) > 1 and not self._serial:
       with Pool() as p:
         headers = p.imap (self._raw_headers, [f for (infile,f) in expanded_infiles])
-        headers = bar.iter(headers)
+        headers = ([h,bar.next()][0] for h in headers)
         headers = list(headers) # Start scanning.
     else:
       headers = imap (self._raw_headers, [f for (infile,f) in expanded_infiles])
-      headers = bar.iter(headers)
+      headers = ([h,bar.next()][0] for h in headers)
       headers = list(headers) # Start scanning.
 
-    bar.finish()
+    # If we made this progress bar, then finish it here.
+    # Otherwise, if it was passed in it could be part of a bigger workflow
+    # so don't finish it yet.
+    if not hasattr(progress,'next'):
+      bar.finish()
 
     # Check which files had headers used, report on the results.
     matches = Counter()
@@ -440,7 +505,7 @@ class BufferBase (object):
       headers = {k:v.copy() for k,v in existing_buffer._headers.items()}
 
     nfiles = len(self._files)
-    if nfiles == 0 and len(infiles) > 0:
+    if nfiles == 0 and existing_buffer is None:
       error(_("no input files found!"))
     elif nfiles > 10:
       global _pandas_needed
@@ -526,7 +591,7 @@ class BufferBase (object):
         # Trim string attributes (remove whitespace padding).
         if isinstance(v,str): v = v.strip()
         # Use regular integers for numeric types.
-        elif np.can_cast(v.dtype,int):
+        elif np.can_cast(v,int):
           v = int(v)
         atts[n] = v
 
@@ -534,10 +599,10 @@ class BufferBase (object):
       axes = OrderedDict()
       for n in self._outer_axes:
         values = var_records[n]
-        # Remove missing values before continuing.
-        values = np.ma.compressed(values)
         # Ignore axes that have no actual coordinate values.
-        if len(values) == 0: continue
+        if len(np.ma.compressed(values)) == 0: continue
+        # Remove missing values before continuing.
+        values = np.ma.unique(values).data
         # Get all unique values (sorted).
         values = tuple(sorted(set(values)))
         if (n,values) not in known_axes:
@@ -605,7 +670,7 @@ class BufferBase (object):
         axes[n] = _dim_type(name=n, length = length)
 
       # Determine the optimal data type to use.
-      dtype_list = var_records['dtype']
+      dtype_list = map(np.dtype,var_records['dtype'])
       dtype = np.result_type(*dtype_list)
 
       var = _iter_type( name = nomvar, atts = atts,
@@ -783,7 +848,7 @@ class BufferBase (object):
         axes[n] = _dim_type(name=n, length = length)
 
       # Determine the optimal data type to use.
-      dtype_list = var_records['dtype']
+      dtype_list = map(np.dtype,var_records['dtype'])
       dtype = np.result_type(*dtype_list)
 
       var = _iter_type( name = nomvar, atts = atts,
@@ -809,6 +874,127 @@ class BufferBase (object):
     else:
       self._makevars_slow()
 
+  # Decode variable structures into a format-independent structure.
+  def _serialize (self):
+    self._makevars()
+    return self._varlist
+
+  # Re-encode format-independent info into a format-specific table.
+  @classmethod
+  def _deserialize (cls, data):
+    # Encapsulate this info in a structure.
+    # Adapted from 'fake_buffer' object created in from_fstpy method.
+    # TODO: simpler way of making empty buffer?
+    b = cls.__new__(cls)
+    b._files = [None]
+    b._varlist = data
+    b._unmakevars()
+    # Check for unused columns.
+    for colname in b._headers.keys():
+      if b._headers._counter[colname] == 0:
+        warn(_("Unknown axis '%s'.  Encoding not complete.")%colname)
+    # Convert headers to regular dictionary.
+    b._headers = dict(b._headers)
+    # Return the result.
+    return b
+
+
+  # Inverse operation - determine record structure from a varlist.
+  def _unmakevars (self):
+    from fstd2nc.mixins import _var_type, _iter_type, _dim_type, _axis_type
+    import numpy as np
+    # By the time we get to this base method, the mixins should have filtered
+    # out all variables that shouldn't be in the record table.
+    varlist = []
+    for var in self._varlist:
+      if isinstance(var,_iter_type):
+        varlist.append(var)
+      else:
+        warn (_("Unable to encode %s.")%var.name)
+    if len(varlist) == 0:
+      error (_("Nothing to encode!"))
+    self._varlist = varlist
+    self._nrecs = sum(np.product(var.record_id.shape) if var.record_id.ndim > 0 else 1 for var in varlist)
+    # Keep track of columns that we don't need to worry about.
+    # (don't care if they get used or not).
+    dontcare = set()
+    # Start constructing header information.
+    # The initial columns will be the outer axes of the variables.
+    # plus the name and data columns, and inner dimensions.
+    headers = dict()
+    for var in self._varlist:
+      # Add outer axes.
+      for axis in var.axes[:var.record_id.ndim]:
+        if axis.name in headers: continue
+        headers[axis.name] = np.ma.masked_all(self._nrecs, dtype=axis.array.dtype)
+      # Add coordinates as well.
+      for coord in var.atts.get('coordinates',[]):
+        if coord.name in headers: continue
+        # It's ok if these values are never actually referenced.
+        dontcare.add(coord.name)
+        headers[coord.name] = np.ma.masked_all(self._nrecs, dtype=coord.array.dtype)
+      # Add inner axes.
+      for dimname in self._inner_axes:
+        headers[dimname] = np.ma.ones(self._nrecs, dtype='int32')
+      for axis in var.axes[var.record_id.ndim:]:
+        if axis.name not in self._inner_axes:
+          error (_("Unhandled inner axis %s.")%axis.name)
+      # Add extra metadata.
+      for attname, attval in var.atts.items():
+        if attname in headers: continue
+        # It's ok if these values are never actually referenced.
+        dontcare.add(attname)
+        # For simple structures, create array with appropriate dtype.
+        # For complicated structures, use object dtype.
+        sample = np.array(attval)
+        if sample.ndim == 0:
+          headers[attname] = np.ma.masked_all(self._nrecs, dtype=sample.dtype)
+        else:
+          headers[attname] = np.ma.masked_all(self._nrecs, dtype=object)
+
+    namesize = max(len(var.name) for var in varlist)
+    headers['name'] = np.ma.masked_all(self._nrecs, dtype='|S%d'%namesize)
+    headers['d'] = np.empty(self._nrecs, dtype=object)
+    headers['file_id'] = np.empty(self._nrecs,dtype='int32')
+    headers['file_id'][:] = -1
+    offset = 0
+    for var in varlist:
+      axes = var.axes[:var.record_id.ndim]
+      # Add outer axes.
+      for i,axis in enumerate(axes):
+        shape = [1]*len(axes)
+        shape[i] = len(axis)
+        array = np.empty(var.record_id.shape, dtype=axis.array.dtype)
+        array[()] = axis.array.reshape(shape)
+        headers[axis.name][offset:offset+var.record_id.size] = array.flatten()
+      # Add coordinate info as well (only if defined on outer axes).
+      for coord in var.atts.get('coordinates',[]):
+        if not all(axis in axes for axis in coord.axes): continue
+        shape = [len(axis) if axis in coord.axes else 1 for axis in axes]
+        array = np.empty(var.record_id.shape, dtype=coord.array.dtype)
+        array[()] = coord.array.reshape(shape)
+        headers[coord.name][offset:offset+var.record_id.size] = array.flatten()
+      # Add inner axes.
+      for axis in var.axes[var.record_id.ndim:]:
+        headers[axis.name][offset:offset+var.record_id.size] = len(axis)
+      # Add extra metadata.
+      for attname, attval in var.atts.items():
+        if attname in ('coordinates',): continue
+        headers[attname][offset:offset+var.record_id.size] = attval
+      headers['name'][offset:offset+var.record_id.size] = var.name
+      headers['d'][offset:offset+var.record_id.size] = var.record_id.flatten()
+      offset = offset + var.record_id.size
+
+    # Use special dictionary to keep track of which columns were used.
+    # It would be a problem if the columns were never processed!
+    self._headers = AccessCountDict(**headers)
+    # Pre-access certain columns, so they don't get flagged as unhandled.
+    # (columns that aren't important).
+    dontcare |= set(('name','d','file_id'))
+    for col in dontcare:
+      self._headers[col]
+    # From here, the mixins will start processing the header info
+
   # Iterate over all unique axes found in the variables.
   # Requires _makevars() to have already been called.
   def _iter_axes (self, name=None, varlist=False):
@@ -886,9 +1072,15 @@ class BufferBase (object):
         yield o
 
   # How to decode the data from a raw binary array.
-  def _decode (self, data, unused):
+  @staticmethod
+  def _decode (data):
     raise NotImplementedError("No decoder found.")
 
+  # Stub for any extra processing needed for the data after it's decoded.
+  @classmethod
+  def _postproc (cls, data):
+    return data
+
   # Shortcuts to header decoding functions.
   # Put into the class so they can potentially be overridden for other formats.
   @staticmethod
@@ -901,11 +1093,35 @@ class BufferBase (object):
   # Shortcut for reading a record, given a record id.
   def _read_record (self, rec):
     import numpy as np
-    with open(self._files[self._headers['file_id'][rec]],'rb') as f:
-      f.seek(self._headers['address'][rec],0)
-      data = np.fromfile(f,'B',self._headers['length'][rec])
-      data = self._decode(data,rec)
-    return data
+    kwargs = {}
+    # Add file-based data.
+    file_id = self._headers['file_id'][rec]
+    if file_id >= 0:
+      f = open(self._files[file_id], 'rb')
+    else:
+      f = None
+    for key, (addr_key, len_key, d_key) in self._decoder_data:
+      if addr_key not in self._headers: continue
+      # Special case: have dask array to read.
+      if d_key in self._headers:
+        d = self._headers[d_key][rec]
+        if d is not None:
+          kwargs[key] = np.asarray(d)
+          continue
+      address = self._headers[addr_key][rec]
+      length = self._headers[len_key][rec]
+      if address == -1 or length == -1: continue
+      f.seek(address,0)
+      data = self._decode(np.fromfile(f,'B',length))
+      kwargs[key] = data
+    for key in self._decoder_extra_args:
+      if key in self._headers:
+        value = self._headers[key][rec]
+        kwargs[key] = value
+    kwargs.update(self._decoder_scalar_args())
+    if f is not None:
+      f.close()
+    return self._postproc(**kwargs)
 
 
   #
