@@ -1,5 +1,5 @@
 ###############################################################################
-# Copyright 2017-2021 - Climate Research Division
+# Copyright 2017-2023 - Climate Research Division
 #                       Environment and Climate Change Canada
 #
 # This file is part of the "fstd2nc" package.
@@ -142,10 +142,8 @@ class FSTD_Compat (BufferBase):
       from os.path import getsize
       import rpnpy.librmn.all as rmn
       # Align the data on 8-byte boundaries, which is what FSTD expects.
-      while True:
-        f.sync()
-        alignment = getsize(filename) % 8
-        if alignment == 0: break
+      f.sync()
+      if getsize(filename) % 8 != 0:
         # Add dummy global attributes as a way to pad out the file.
         # The chosen size for the attributes is from trial-and-error. It seems
         # to get the right padding after a few iterations.
@@ -156,7 +154,14 @@ class FSTD_Compat (BufferBase):
         # The current approach of writing attributes wastes more space (a few
         # dozen kilobytes?) but it seems to do the job.
         #f.setncattr('_pad%05d'%npad[0], np.zeros(4096+(8-alignment),dtype='B'))
-        f.setncattr('_pad%05d'%npad[0], np.zeros(0+(npad[0]%7),dtype='B'))
+        #f.setncattr('_pad%05d'%npad[0], np.zeros(0+(npad[0]%7),dtype='B'))
+        padding = 1
+        while True:
+          f.setncattr('_pad%05d'%npad[0], np.zeros(padding,dtype='B'))
+          f.sync()
+          if getsize(filename) % 8 == 0: break
+          f.delncattr('_pad%05d'%npad[0])
+          padding += 127
         npad[0] += 1
       # Next, write the data
       address = getsize(filename)
@@ -205,10 +210,10 @@ class FSTD_Compat (BufferBase):
         continue
       # Hard case: only have the record indices, need to loop over the records.
       # Get the shape of a single record for the variable.
-      if hasattr(var,'record_id'):
+      if hasattr(var, 'chunks'):
+        continue  #TODO
+      elif hasattr(var,'record_id'):
         record_shape = var.shape[var.record_id.ndim:]
-      elif hasattr(var,'chunksize'):
-        record_shape = var.chunksize
       else:
         continue
       # Use this as the "chunk size" for the netCDF file, to improve I/O
@@ -367,7 +372,9 @@ class FSTD_Compat (BufferBase):
     # nj, datyp
     buf[:,4] = (headers['nj']<<8) + datyps
     # nk, ubc
-    buf[:,5] = (headers['nk']<<12) + headers['ubc']
+    # (nk and ubc no longer guaranteed to be kept in the headers).
+    #buf[:,5] = (headers['nk']<<12) + headers['ubc']
+    buf[:,5] = (1<<12) + 0
     # npas, pad7
     buf[:,6] = (headers['npas']<<6)
     # ig4, ig2a
