@@ -1,5 +1,5 @@
 ###############################################################################
-# Copyright 2017-2021 - Climate Research Division
+# Copyright 2017-2023 - Climate Research Division
 #                       Environment and Climate Change Canada
 #
 # This file is part of the "fstd2nc" package.
@@ -29,6 +29,12 @@ from fstd2nc.mixins import BufferBase
 # Remove extraneous dimensions from the output.
 
 class NoNK (BufferBase):
+  def __init__ (self, *args, **kwargs):
+    import numpy as np
+    super(NoNK,self).__init__(*args, **kwargs)
+    if 'nk' in self._headers:
+      if np.all(self._headers['nk'] == 1):
+        del self._headers['nk']
   def _makevars (self):
     super(NoNK,self)._makevars()
     for var in self._varlist:
