@@ -1,5 +1,5 @@
 ###############################################################################
-# Copyright 2017-2021 - Climate Research Division
+# Copyright 2017-2023 - Climate Research Division
 #                       Environment and Climate Change Canada
 #
 # This file is part of the "fstd2nc" package.
@@ -79,6 +79,7 @@ class netCDF_Atts (BufferBase):
       import configparser as ConfigParser
     from collections import OrderedDict
     import numpy as np
+    from os.path import exists
     metadata_file = kwargs.pop('metadata_file',None)
     if metadata_file is None:
       metafiles = []
@@ -95,9 +96,15 @@ class netCDF_Atts (BufferBase):
       metadata['global'] = OrderedDict(Conventions = conventions)
 
     # Read the metadata files.
-    configparser = ConfigParser.SafeConfigParser()
+    try:
+      configparser = ConfigParser.SafeConfigParser()
+    # Python 3.12+ removed SafeConfigParser, so use ConfigParser instead.
+    except AttributeError:
+      configparser = ConfigParser.ConfigParser()
     configparser.optionxform = str # Make the attribute names case sensitive.
     for metafile in metafiles:
+      if not exists(metafile):
+        error(_("'%s' does not exist.")%metafile)
       configparser.read(metafile,encoding='utf-8')
     for varname in configparser.sections():
       metadata.setdefault(varname,OrderedDict()).update(configparser.items(varname))
@@ -261,8 +268,8 @@ class netCDF_IO (BufferBase):
 
   def _fix_names (self):
 
-    # List of metadata keys that are internal to the FSTD file.
-    internal_meta = self._headers.keys()
+    # List of metadata keys that are internal to the input file.
+    internal_meta = list(self._headers.keys()) + list(self._ignore_atts)
 
     # Generate unique axis names.
     axis_table = dict()
@@ -399,10 +406,10 @@ class netCDF_IO (BufferBase):
         continue
       # Hard case: only have the record indices, need to loop over the records.
       # Get the shape of a single record for the variable.
-      if hasattr(var,'record_id'):
+      if hasattr(var, 'chunks'):
+        continue  #TODO
+      elif hasattr(var,'record_id'):
         record_shape = var.shape[var.record_id.ndim:]
-      elif hasattr(var,'chunksize'):
-        record_shape = var.chunksize
       else:
         continue
       # Use this as the "chunk size" for the netCDF file, to improve I/O
@@ -450,17 +457,15 @@ class netCDF_IO (BufferBase):
       import numpy as np
       io = sorted(io)
       with Pool() as p:
-        raw = p.imap (_quick_load, ((self._files[self._headers['file_id'][r]], self._headers['address'][r], self._headers['length'][r]) for r,shape,v,ind in io))
+        raw = p.imap (self._read_record, (r for r,shape,v,ind in io))
         raw = bar.iter(raw)
-        for (r,shape,v,ind), stuff in zip(io, raw):
-          data = self._decode (stuff, r)
+        for (r,shape,v,ind), data in zip(io, raw):
           v[ind] = data.astype(v.dtype).reshape(shape)
         bar.finish()
     else:
       for r,shape,v,ind in bar.iter(sorted(io)):
         try:
-          stuff = _quick_load ((self._files[self._headers['file_id'][r]], self._headers['address'][r], self._headers['length'][r]))
-          data = self._decode (stuff, r)
+          data = self._read_record (r)
           v[ind] = data.astype(v.dtype).reshape(shape)
         except (IndexError,ValueError):
           warn(_("Internal problem with the script - unable to get data for '%s'")%v.name)
@@ -471,20 +476,30 @@ class netCDF_IO (BufferBase):
   # Alias "to_netcdf" as "write_nc_file" for backwards compatibility.
   write_nc_file = to_netcdf
 
-# Internal helper method for delegating the load to a multiprocessing Pool.
-def _quick_load (args):
-  import numpy as np
-  filename, address, length = args
-  # Check if this was from an external source, in which case we have no file
-  # data to provide.  The _decode method will have to find the data from the
-  # other argument (record index).
-  if address is None or length is None:
-    # Return some minimal amount of empty data, in case a mixin tries to
-    # access this data.  This is probably a bad idea (in case the mixin
-    # does something weird with the zero values).
-    #TODO: fix the mixins so they don't try to read these values directly.
-    return np.zeros(256,'B')
-  with open (filename, 'rb') as f:
-    f.seek (address, 0)
-    return np.fromfile(f,'B',length)
+  #TODO
+  # Attempt to undo the uniquification of axis/variable names, such as what
+  # was caused by _fix_names.
+  # Mainly useful for going in reverse (netcdf back to fstd).
+  def _unfix_names (self):
+    # Helper method - strip integer indices from dimension names.
+    def strip_numbers (name):
+      while name[-1] in '_0123456789':
+        name = name[:-1]
+      return name
+    for axis in self._iter_axes():
+      axis.name = strip_numbers(axis.name)
+    for coord in self._iter_coords():
+      coord.name = strip_numbers(coord.name)
+    for var in self._varlist:
+      # For 1D or 0D variables, strip out underscores and numbers.
+      # Assume such variables were probably meant to be coordinates, but were
+      # not in the 'coords' attribute for some particular reason
+      # (such as for leadtime, reftime).
+      if len(var.axes) <= 1:
+        var.name = strip_numbers(var.name)
+
+  # Re-encode variables with netCDF modifications undone.
+  def _unmakevars (self):
+    self._unfix_names()
+    super(netCDF_IO,self)._unmakevars()
 
