@@ -1,5 +1,5 @@
 ###############################################################################
-# Copyright 2017-2021 - Climate Research Division
+# Copyright 2017-2023 - Climate Research Division
 #                       Environment and Climate Change Canada
 #
 # This file is part of the "fstd2nc" package.
@@ -52,6 +52,38 @@ def decode_ip1_level (ip1):
   else:
     return r1.v1
 
+# Internal helper function - take packed (kind,level) info and convert to
+# ip1 value.
+@vectorize
+def _encode_ip1 (kind_level):
+  import rpnpy.librmn.all as rmn
+  import numpy as np
+  # Decode kind and level information (was packed in single 64-bit value).
+  kind_level = np.array([kind_level],dtype='uint64').view([('kind','int32'),('level','float32')])
+  kind = int(kind_level['kind'][0])
+  level = float(kind_level['level'][0])
+  rip1 = rmn.FLOAT_IP(level,level,kind)
+  rip2 = rmn.FLOAT_IP(0,0,rmn.KIND_HOURS)
+  rip3 = rmn.FLOAT_IP(0,0,rmn.KIND_ARBITRARY)
+  ip1, ip2, ip3 = rmn.EncodeIp(rip1,rip2,rip3)
+  return ip1
+
+# Helper function - take arrays of kind, level and encode as array of ip1 values.
+def encode_ip1 (kind, level):
+  import numpy as np
+  mask = None
+  if hasattr(level,'mask'):
+    mask = np.ma.getmaskarray(level)
+    level = level.filled(0.0)
+  # Hack the kind and level information into something that's hashable.
+  kind_level = np.empty(kind.shape, dtype=[('kind','int32'),('level','float32')])
+  kind_level['kind'] = kind
+  kind_level['level'] = level
+  kind_level = kind_level.view('uint64')
+  if mask is not None:
+    return np.ma.masked_array(_encode_ip1(kind_level), mask=mask)
+  else:
+    return np.array(_encode_ip1(kind_level))
 
 
 #################################################
@@ -72,6 +104,14 @@ class VCoords (BufferBase):
     parser.add_argument('--momentum-levels', '--mlev', action='store_true', help=_("Only convert data that's on 'momentum' vertical levels."))
     parser.add_argument('--vertical-velocity-levels', '--wlev', action='store_true', help=_("Only convert data that's on 'vertical velocity' levels."))
 
+  # Tell the decoder not to process vertical records as variables.
+  @classmethod
+  def _meta_records(cls):
+    return super(VCoords,cls)._meta_records() + (b'!!',b'!!SF')
+  @classmethod
+  def _maybe_meta_records(cls):
+    return super(VCoords,cls)._maybe_meta_records() + (b'HY',)
+
   def __init__ (self, *args, **kwargs):
     """
     strict_vcoord_match : bool, optional
@@ -114,9 +154,7 @@ class VCoords (BufferBase):
 
     # Use decoded IP1 values as the vertical axis.
     self._outer_axes = ('level',) + self._outer_axes
-    # Tell the decoder not to process vertical records as variables.
-    self._meta_records = self._meta_records + (b'!!',b'!!SF')
-    self._maybe_meta_records = self._maybe_meta_records + (b'HY',)
+    self._ignore_atts = ('level_descr','special_level') + self._ignore_atts
     super(VCoords,self).__init__(*args,**kwargs)
     # Don't group records across different level 'kind'.
     # (otherwise can't create a coherent vertical axis).
@@ -128,9 +166,11 @@ class VCoords (BufferBase):
     # to ignore them).
     vrecs = OrderedDict()
     # Keep track of any diagnostic levels, e.g. from GEM.
-    diag_ip1 = []
+    self._diag_ip1 = []
     for vcoord_nomvar in (b'HY  ',b'!!  '):
-      for handle in self._fstinl(nomvar=vcoord_nomvar):
+      # Note: assume there will only be a few unique records, so limit the
+      # processing.
+      for handle in self._fstinl(nomvar=vcoord_nomvar)[:10]:
         header = self._fstprm(handle)
         key = (header['ip1'],header['ip2'])
         # For old HY records, there's no matching ipX/igX codes.
@@ -146,9 +186,9 @@ class VCoords (BufferBase):
           # Problems opening this vertical coordinate?
           continue
         try:
-          diag_ip1.append(vgd_get(vgd_id,'DIPT'))
-          diag_ip1.append(vgd_get(vgd_id,'DIPM'))
-          diag_ip1.append(vgd_get(vgd_id,'DIPW'))
+          self._diag_ip1.append(vgd_get(vgd_id,'DIPT'))
+          self._diag_ip1.append(vgd_get(vgd_id,'DIPM'))
+          self._diag_ip1.append(vgd_get(vgd_id,'DIPW'))
         except (KeyError,VGDError):
           #warn(_("Unable to parse diagnostic levels from the vertical coordinate"))
           pass
@@ -176,13 +216,13 @@ class VCoords (BufferBase):
     fields = self._headers
     # Provide 'level' and 'kind' information to the decoder.
     fields['level'] = np.array(decode_ip1_level(fields['ip1']), dtype='float32')
-    fields['kind'] = np.array(decode_ip1_kind(fields['ip1']), dtype='int32')
+    fields['kind'] = np.array(decode_ip1_kind(fields['ip1']), dtype='uint8')
 
     # Pre-filter the diagnostic level data?
     # Start by treating diagnostic level as model level, then
     # revert this later if it ends up not working.
     if not self._split_diag_level and not self._only_diag_level:
-      for ip1_val in diag_ip1:
+      for ip1_val in self._diag_ip1:
           mask = (fields['ip1'] == ip1_val)
           if self._ignore_diag_level:
             fields['selected'] &= (~mask)
@@ -194,7 +234,7 @@ class VCoords (BufferBase):
     # Only use diagnostic level(s)?
     if self._only_diag_level:
       mask = False
-      for ip1_val in diag_ip1:
+      for ip1_val in self._diag_ip1:
           mask |= (fields['ip1'] == ip1_val)
       fields['selected'] &= mask
 
@@ -244,37 +284,46 @@ class VCoords (BufferBase):
           warn(_("Unable to parse vertical velocity levels from the vertical coordinate"))
         vgd_free (vgd_id)
 
-    # Apply special labels for some levels.
-    # Used to describe a variable when it's split into multiple coordinate types.
-    fields['level_descr'] = np.empty(len(fields['kind']), dtype='|S20')
-    fields['level_descr'][:] = 'unknown_levels'
-    fields['level_descr'][fields['kind'] == 0] = 'depth_levels'
-    fields['level_descr'][fields['kind'] == 1] = 'model_levels'
-    fields['level_descr'][fields['kind'] == 2] = 'pressure_levels'
-    fields['level_descr'][fields['kind'] == 3] = 'generic_levels'
-    fields['level_descr'][fields['kind'] == 4] = 'height_levels'
-    fields['level_descr'][fields['kind'] == 5] = 'model_levels'
-    fields['level_descr'][fields['kind'] == 21] = 'model_levels'
-    fields['special_level'] = np.empty(len(fields['kind']), dtype='|S20')
-    fields['special_level'][:] = fields['level_descr'][:]
-    # Discern ocean depths from ocean bottom.
-    fields['special_level'][fields['kind'] == 0] = 'depth_levels'
-    fields['special_level'][(fields['kind'] == 1) & (fields['level'] == 1.0)] = 'bottom_level'
-    # Discern atmospheric levels from diagnostic levels.
-    for ip1_val in diag_ip1:
-      fields['special_level'][fields['ip1'] == ip1_val] = 'diag_level'
+
+  # Add human readable vcoord attributes for making unique variable names.
+  def _insert_vcoord_atts (self):
+    for var in self._varlist:
+      if 'kind' not in var.atts: continue
+      kind = var.atts['kind']
+      # Apply special labels for some levels.
+      # Used to describe a variable when it's split into multiple coordinate types.
+      var.atts['level_descr'] = 'unknown_levels'
+      if kind == 0: var.atts['level_descr'] = 'depth_levels'
+      elif kind == 1: var.atts['level_descr'] = 'model_levels'
+      elif kind == 2: var.atts['level_descr'] = 'pressure_levels'
+      elif kind == 3: var.atts['level_descr'] = 'generic_levels'
+      elif kind == 4: var.atts['level_descr'] = 'height_levels'
+      elif kind == 5: var.atts['level_descr'] = 'model_levels'
+      elif kind == 21: var.atts['level_descr'] = 'model_levels'
+      var.atts['special_level'] = var.atts['level_descr']
+      # Discern ocean depths from ocean bottom.
+      if kind == 0: var.atts['special_level'] = 'depth_levels'
+      elif kind == 1 and var.atts.get('ip1',None) == 26314400:
+        var.atts['special_level'] = 'bottom_level'
+      # Discern atmospheric levels from diagnostic levels.
+      if var.atts.get('ip1',None) in self._diag_ip1:
+        var.atts['special_level'] = 'diag_level'
 
   def _makevars (self):
     super(VCoords,self)._makevars()
+
+    self._insert_vcoord_atts()
+
     # Call this makevars routine from a separate wrapper, to make it easier
     # to re-run just that one part without redoing the whole pipeline.
     # I.e., for case where our initial attempt at fitting the diagnostic level
     # into the model levels fails.
     try:
-      return self._vcoords_makevars()
+      self._vcoords_makevars()
     except ValueError:
       super(VCoords,self)._makevars()
-      return self._vcoords_makevars()
+      self._insert_vcoord_atts()
+      self._vcoords_makevars()
 
   def _vcoords_makevars (self):
     from fstd2nc.mixins import _var_type, _axis_type, _dim_type
@@ -283,6 +332,7 @@ class VCoords (BufferBase):
     from rpnpy.vgd.base import vgd_fromlist, vgd_get, vgd_free
     from rpnpy.vgd.const import VGD_KEYS, VGD_OPR_KEYS
     from rpnpy.vgd import VGDError
+    import rpnpy.librmn.all as rmn
 
     vrecs = self._vrecs
 
@@ -290,6 +340,12 @@ class VCoords (BufferBase):
     # again with updated information.
     rerun = False
 
+    # Check if cell boundaries are requested.
+    # This is controlled from the xycoords mixin, but we can also add
+    # vertical bounds as well.
+    bounds = self._bounds
+    from fstd2nc.mixins.xycoords import bnds2
+
     # Scan through the data, and look for any use of vertical coordinates.
     vaxes = OrderedDict()
     prefs = dict()  # Reference pressure scalar variables.
@@ -343,6 +399,7 @@ class VCoords (BufferBase):
         atts = OrderedDict()
         name = 'level'
         new_axis = _axis_type(name, atts, level_axis.array)
+        new_bnds = None
 
         # Check if we have a vertical coordinate record to use.
         if key in vrecs:
@@ -367,6 +424,10 @@ class VCoords (BufferBase):
             coordA = _var_type('a', {}, [new_axis], None)
             coordB = _var_type('b', {}, [new_axis], None)
             coordC = _var_type('c', {}, [new_axis], None)
+            # Partial definition of a/b boundaries.
+            coordA_bnds = _var_type('a_bnds', {}, [new_axis,bnds2], None)
+            coordB_bnds = _var_type('b_bnds', {}, [new_axis,bnds2], None)
+            coordC_bnds = _var_type('c_bnds', {}, [new_axis,bnds2], None)
 
             # First, patch up VGD_OPR_KEYS to include SLEVE coordinates?
             #TODO: remove this once rpnpy has full support for this.
@@ -385,7 +446,18 @@ class VCoords (BufferBase):
             # Add all parameters for this coordinate.
             for vgdkey in VGD_KEYS:
               try:
+                # Recent versions of vgrid are very verbose about querying for
+                # attributes that don't exist.  Turn off those messages if
+                # the App logging mechanism is available.
+                # If App logging is not available, then vgrid is not recent
+                # enough to have this verbosity problem anyway.
+                try:
+                  loglvl = rmn.librmn.Lib_LogLevelNo(5,9)
+                except AttributeError: pass
                 val = vgd_get(vgd_id,vgdkey)
+                try:
+                  rmn.librmn.Lib_LogLevelNo(5,loglvl)
+                except AttributeError: pass
                 # Skip multidimensional arrays (can't be encoded as metadata).
                 if getattr(val,'ndim',1) > 1: continue
                 internal_atts[vgdkey] = val
@@ -432,11 +504,63 @@ class VCoords (BufferBase):
               coordA.array = np.asarray(A)
               coordB.array = np.asarray(B)
               coordC.array = np.asarray(C)
+              # Find vertical layer boundaries.
+              if bounds:
+                A_bnds1, A_bnds2 = [], []
+                B_bnds1, B_bnds2 = [], []
+                C_bnds1, C_bnds2 = [], []
+                z_bnds1, z_bnds2 = [], []
+                all_z_sorted = sorted(set(all_z))
+                for z in levels:
+                  ind = all_z_sorted.index(z)
+                  # Catch values out of range.
+                  # This was triggered when run on field that only had
+                  # diagnostic level.  In that case, put any placeholder value
+                  # assuming it won't be used in the end (wrong coordinate type
+                  # for a/b values to show up in the output anyway).
+                  if ind > 0:
+                    z1 = all_z_sorted[ind-1]
+                    ind1 = all_z.index(z1)
+                    A_bnds1.append(all_a[ind1])
+                    B_bnds1.append(all_b[ind1])
+                    z_bnds1.append(all_z[ind1])
+                    if sleve:
+                      C_bnds1.append(all_c[ind1])
+                  else:
+                    A_bnds1.append(float('nan'))
+                    B_bnds1.append(float('nan'))
+                    z_bnds1.append(float('nan'))
+                    if sleve:
+                      C_bnds1.append(float('nan'))
+                  if ind < len(all_z_sorted)-1:
+                    z2 = all_z_sorted[ind+1]
+                    ind2 = all_z.index(z2)
+                    A_bnds2.append(all_a[ind2])
+                    B_bnds2.append(all_b[ind2])
+                    z_bnds2.append(all_z[ind2])
+                    if sleve:
+                      C_bnds2.append(all_c[ind2])
+                  else:
+                    A_bnds2.append(float('nan'))
+                    B_bnds2.append(float('nan'))
+                    z_bnds2.append(float('nan'))
+                    if sleve:
+                      C_bnds2.append(float('nan'))
+                coordA_bnds.array = np.array(np.asarray([A_bnds1,A_bnds2]).T)
+                coordB_bnds.array = np.array(np.asarray([B_bnds1,B_bnds2]).T)
+                new_bnds = _var_type(name+'_bnds', {}, [new_axis,bnds2], None)
+                new_bnds.array = np.array(np.asarray([z_bnds1,z_bnds2]).T)
+                if sleve:
+                  coordC_bnds.array = np.array(np.asarray([C_bnds1,C_bnds2]).T)
             except (KeyError,ValueError,VGDError):
               warn (_("Unable to get A/B coefficients for %s.")%var.name)
               coordA = None
               coordB = None
               coordC = None
+              coordA_bnds = None
+              coordB_bnds = None
+              coordC_bnds = None
+              new_bnds = None
             vgd_free (vgd_id)
         ###
 
@@ -519,6 +643,7 @@ class VCoords (BufferBase):
                 atts['standard_name'] = 'atmosphere_hybrid_sigma_pressure_coordinate'
                 if None not in (coordA,coordB):
                   coordA.array /= 100.0  # Use hPa for final units.
+                  if coordA_bnds is not None and coordA_bnds.array is not None: coordA_bnds.array /= 100.0
                   atts['formula'] = 'p = ap + b*ps'
                   coordA.name = 'ap'
                   atts['formula_terms'] = OrderedDict([('ap',coordA),('b',coordB),('ps','P0')])
@@ -575,6 +700,25 @@ class VCoords (BufferBase):
               atts['formula_terms'] = OrderedDict([('a',coordA),('b',coordB),('orog','ME')])
               coordinates.extend([coordA,coordB])
 
+        # Attach vertical boundaries?
+        if bounds and new_bnds is not None and np.all(np.isfinite(new_bnds.array)):
+          atts['bounds'] = new_bnds
+          # Attach a/b bounds
+          if 'formula_terms' in atts:
+            new_bnds.atts['formula_terms'] = atts['formula_terms'].copy()
+            for k, v in new_bnds.atts['formula_terms'].items():
+              if v is coordA:
+                new_bnds.atts['formula_terms'][k] = coordA_bnds
+                coordA.atts['bounds'] = coordA_bnds
+                coordA_bnds.name = coordA.name+'_bnds'
+              if v is coordB:
+                new_bnds.atts['formula_terms'][k] = coordB_bnds
+                coordB.atts['bounds'] = coordB_bnds
+                coordB_bnds.name = coordB.name+'_bnds'
+              if v is coordC:
+                new_bnds.atts['formula_terms'][k] = coordC_bnds
+                coordC.atts['bounds'] = coordC_bnds
+                coordC_bnds.name = coordC.name+'_bnds'
         # Add this vertical axis.
         if len(coordinates) > 0:
           atts['coordinates'] = coordinates
@@ -634,3 +778,129 @@ class VCoords (BufferBase):
 
     if rerun:
       raise ValueError
+
+  # Re-encode vertical coordinate info back into FSTD metadata.
+  def _unmakevars (self):
+    from fstd2nc.mixins import _iter_type, _var_type, _axis_type, _dim_type
+    import numpy as np
+    import rpnpy.vgd.all as vgd
+
+    # Helper - add a coordinate record
+    def add_coord (name,nj,ni,values,**atts):
+      import dask.array as da
+      dims = []
+      if nj > 1: dims.append(_dim_type('j',nj))
+      if ni > 1: dims.append(_dim_type('i',ni))
+      # Need to carefully encode the arrays so they are scalar object containing the data.
+      array = np.empty(1,dtype=object)
+      array[0] = da.from_array(values, chunks=-1)
+      array = array.squeeze()
+      self._varlist.append(_iter_type(name,atts,dims,'float32',array))
+
+    axis_map = dict()
+    # Detect special case where depth could be encoded as height
+    depth_as_height = any(var.atts.get('typvar',None) == 'P@' for var in self._varlist)
+    # Look for vertical coordinates, annotate accordingly.
+    for axis in self._iter_axes():
+      # Skip dimensions that have no coordinate system / attributes
+      # (those are not going to be vertical levels).
+      if not hasattr(axis,'atts'): continue
+      standard_name = axis.atts.get('standard_name',None)
+      kind = None
+      # If height coordinate, defer to kind=4 for special case of diagnostic level.
+      if (standard_name == 'height' and (len(axis) > 1 or getattr(axis,'array',None) not in [1.5,10.0])) or (standard_name == 'depth' and depth_as_height):
+        kind = 0
+      elif standard_name == 'atmosphere_sigma_coordinate':
+        kind = 1
+      elif standard_name == 'air_pressure' or axis.name == 'pres' or axis.atts.get('units',None) == 'hPa':
+        #TODO: unit conversion from Pa?
+        kind = 2
+      elif standard_name == 'model_level_number':
+        kind = 3
+      elif standard_name == 'height' or axis.name == 'height':
+        kind = 4
+      elif standard_name == 'atmosphere_hybrid_sigma_ln_pressure_coordinate':
+        kind = 5
+      elif standard_name == 'atmosphere_hybrid_sigma_pressure_coordinate':
+        kind = 5
+      elif standard_name == 'air_potential_temperature':
+        kind = 6
+      elif standard_name == 'depth' or axis.name == 'depth':
+        kind = 7
+      elif standard_name == 'atmosphere_sleve_coordinate':
+        kind = 21
+      elif standard_name == 'atmosphere_hybrid_height_coordinate':
+        kind = 21
+      elif axis.atts.get('axis',None) == 'Z' or axis.name == 'level':
+        kind = 3  # Arbitrary level?
+      # Nothing to do for non-vertical axes.
+      if kind is None: continue
+
+      atts = axis.atts.copy()
+      # Add the detected vertical coordinate kind.
+      # Overwrite any exising 'KIND', which may be conflicting.
+      # (e.g., for diagnostic level, still have KIND 5 which causes problems
+      # with ip1 encoding).
+      atts['KIND'] = kind
+
+      # Map to a generic 'level' axis which will be transformed into a 'level' column in the table.
+      axis_map[id(axis)] = _axis_type('level',atts,axis.array)
+
+      # Create vertical coordinate record.
+      try:
+        if atts['KIND'] in (5,21):
+          if atts.get('RC_3',0) < 0: del atts['RC_3']
+          if atts.get('RC_4',0) < 0: del atts['RC_4']
+          if 'DHM' not in atts:
+            atts['DHM'] = decode_ip1_level(atts['DIPM'])
+          if 'DHT' not in atts:
+            atts['DHT'] = decode_ip1_level(atts['DIPT'])
+          vid = vgd.vgd_new2 (
+            kind=atts['KIND'], version=atts['VERS'], hyb=axis.array,
+            rcoef1=atts['RC_1'], rcoef2=atts['RC_2'],
+            rcoef3=atts.get('RC_3',None), rcoef4=atts.get('RC_4',None),
+            pref=atts['PREF'], ptop=atts.get('PTOP',None),
+            dhm=atts['DHM'], dht=atts['DHT']
+          )
+          table = vgd.vgd_tolist(vid).squeeze().T
+          # Add a record for each matching ig1/ig2/ig3.
+          handled_toctocs = set()
+          for var in self._varlist:
+            if not axis in var.axes: continue
+            ig1 = var.atts['ig1']
+            ig2 = var.atts['ig2']
+            ig3 = var.atts['ig3']
+            if (ig1,ig2,ig3) in handled_toctocs: continue
+            add_coord ('!!',table.shape[0],table.shape[1],table,datyp=5,nbits=64,typvar='X',grtyp='X',ip1=ig1,ip2=ig2,ip3=ig3,ig1=atts['KIND']*1000+atts['VERS'])
+            handled_toctocs.add((ig1,ig2,ig3))
+      except KeyError:
+        pass
+
+    # Update the vertical axis to the generic one.
+    for var in self._varlist:
+      for i,axis in enumerate(var.axes):
+        if id(axis) in axis_map:
+          axis = axis_map[id(axis)]
+          var.axes[i] = axis
+          if 'KIND' in axis.atts:
+            var.atts['kind'] = axis.atts['KIND']
+
+    # Strip off any suffixes added to the variable name that are coming from
+    # the vertical axis type (e.g. _model_levels, _diag_level)
+    for var in self._varlist:
+      if var.name.endswith('_level') or var.name.endswith('_levels'):
+        var.name = '_'.join(var.name.split('_')[:-2])
+
+    super(VCoords,self)._unmakevars()
+
+    # Generate ip1 values from level, kind pairs.
+    if 'ip1' not in self._headers.keys():
+      self._headers['ip1'] = np.ma.masked_all(self._nrecs,dtype='int32')
+    if 'kind' not in self._headers.keys():  # Force generic level type?
+      self._headers['kind'] = np.ma.masked_all(self._nrecs,dtype='int32')
+    if 'level' in self._headers.keys():
+      # Skip encoding ip1 if it's already been encoded elsewhere.
+      have_level = ~np.ma.getmaskarray(self._headers['level'])
+      self._headers['ip1'][have_level] = encode_ip1(self._headers['kind'],self._headers['level'])[have_level]
+    # Use default value where ip1 is not used.
+    self._headers['ip1'] = self._headers['ip1'].filled(0)
